import os
import shutil
import sys

import click

from biolib._internal import llm_instructions


@click.group(name='sdk', help='Advanced commands for developers')
def sdk():
    pass


@sdk.command(
    name='add-copilot-prompts', help='Add BioLib-specific GitHub Copilot prompts and instructions to your repository'
)
@click.option('--force', is_flag=True, help='Force overwrite existing files.')
def add_copilot_prompts(force: bool) -> None:
    current_working_directory = os.getcwd()
    config_file_path = f'{current_working_directory}/.biolib/config.yml'
    if not os.path.exists(config_file_path):
        err_string = """
Error: Current directory has not been initialized as a BioLib application.
       Please run the "biolib init" command first"""
        click.echo(err_string, file=sys.stderr)
        exit(1)
    source_path = os.path.join(os.path.dirname(llm_instructions.__file__), '.github')
    destination_path = os.path.join(current_working_directory, '.github')

    conflicting_files = []

    for root, _, files in os.walk(source_path):
        relative_dir = os.path.relpath(root, source_path)
        destination_dir = os.path.join(destination_path, relative_dir)
        for file in files:
            source_file = os.path.join(root, file)
            destination_file = os.path.join(destination_dir, file)
            if os.path.exists(destination_file) and not force:
                with open(source_file, 'rb') as fsrc, open(destination_file, 'rb') as fdest:
                    if fsrc.read() != fdest.read():
                        conflicting_files.append(os.path.relpath(destination_file, current_working_directory))
            else:
                os.makedirs(destination_dir, exist_ok=True)
                shutil.copy2(source_file, destination_file)

    if conflicting_files:
        click.echo('The following files were not overwritten. Use --force to override them:', file=sys.stderr)
        for conflicting_file in conflicting_files:
            click.echo(f'  {conflicting_file}', file=sys.stderr)
        exit(1)

    click.echo(f'Prompt and instruction files added to {destination_path}/')



import time
import uuid


def open_browser_window_from_notebook(url_to_open: str) -> None:
    try:
        from IPython.display import (  # type:ignore # pylint: disable=import-error, import-outside-toplevel
            Javascript,
            display,
            update_display,
        )
    except ImportError as error:
        raise Exception('Unexpected environment. This function can only be called from a notebook.') from error

    display_id = str(uuid.uuid4())
    display(Javascript(f'window.open("{url_to_open}");'), display_id=display_id)
    time.sleep(1)
    update_display(Javascript(''), display_id=display_id)



import logging
import sys

import click

from biolib import api, biolib_errors
from biolib.biolib_api_client.api_client import BiolibApiClient
from biolib.biolib_logging import logger, logger_no_user_data
from biolib.user import sign_in, sign_out


@click.command(help='Login your to BioLib account with web browser')
@click.option(
    '-w',
    is_flag=True,
    default=False,
    required=False,
    type=bool,
    help='Automatically open the login page in the default web browser',
)
def login(w: bool) -> None:  # pylint: disable=invalid-name
    logger.configure(default_log_level=logging.INFO)
    logger_no_user_data.configure(default_log_level=logging.INFO)
    sign_in(open_in_default_browser=w)


@click.command(help='Logout of your BioLib account')
def logout() -> None:
    logger.configure(default_log_level=logging.INFO)
    logger_no_user_data.configure(default_log_level=logging.INFO)
    sign_out()


@click.command(help='Prints out the full name of the user logged in')
def whoami() -> None:
    client = BiolibApiClient.get()
    if client.is_signed_in:
        user_uuid = None
        if client.access_token is None:
            print('Unable to fetch user credentials. Please try logging out and logging in again.')
            exit(1)
        try:
            user_uuid = client.decode_jwt_without_checking_signature(jwt=client.access_token)['payload']['public_id']
        except biolib_errors.BioLibError as error:
            print(
                f'Unable to reference user public_id in access token:\n {error.message}',
                file=sys.stderr,
            )
            exit(1)
        response = api.client.get(path=f'/user/{user_uuid}/')
        user_dict = response.json()
        email = user_dict['email']
        intrinsic_account = [account for account in user_dict['accounts'] if account['role'] == 'intrinsic'][0]
        display_name = intrinsic_account['display_name']
        print(f'Name: {display_name}\nEmail: {email}\nLogged into: {client.base_url}')
    else:
        print('Not logged in', file=sys.stderr)
        exit(1)



from typing import Optional

# Imports to hide and use as private internal utils
from biolib._data_record.data_record import DataRecord as _DataRecord
from biolib._internal.push_application import push_application as _push_application
from biolib._internal.push_application import set_app_version_as_active as _set_app_version_as_active
from biolib._runtime.runtime import Runtime as _Runtime
from biolib._session.session import Session as _Session
from biolib.app import BioLibApp as _BioLibApp

# Classes to expose as public API
Runtime = _Runtime


def get_session(refresh_token: str, base_url: Optional[str] = None) -> _Session:
    return _Session.get_session(refresh_token=refresh_token, base_url=base_url)


def push_app_version(uri: str, path: str) -> _BioLibApp:
    push_data = _push_application(
        app_uri=uri,
        app_path=path,
        app_version_to_copy_images_from=None,
        set_as_active=False,
        set_as_published=False,
    )
    uri = f'{push_data["app_uri"]}:{push_data["sematic_version"]}'
    return _BioLibApp(uri)


def set_app_version_as_default(app_version: _BioLibApp) -> None:
    app_version_uuid = app_version.version['public_id']
    _set_app_version_as_active(app_version_uuid)


def get_app_version_pytest_plugin(app_version: _BioLibApp):
    try:
        import pytest  # type: ignore # pylint: disable=import-outside-toplevel,import-error
    except BaseException:
        raise Exception('Failed to import pytest; please make sure it is installed') from None

    class AppVersionFixturePlugin:
        def __init__(self, app_version_ref):
            self.app_version_ref = app_version_ref

        @pytest.fixture(scope='session')
        def app_version(self, request):  # pylint: disable=unused-argument
            return self.app_version_ref

    return AppVersionFixturePlugin(app_version)


def create_data_record(
    destination: str,
    data_path: str,
    name: Optional[str] = None,
    record_type: Optional[str] = None,
) -> _DataRecord:
    return _DataRecord.create(
        destination=f'{destination}/{name}' if name else destination,
        data_path=data_path,
        record_type=record_type,
    )



from biolib.typing_utils import TypedDict


class RuntimeJobDataDict(TypedDict):
    version: str
    job_requested_machine: str
    job_uuid: str
    job_auth_token: str
    app_uri: str
    is_environment_biolib_cloud: bool


class BioLibRuntimeError(Exception):
    pass


class BioLibRuntimeNotRecognizedError(BioLibRuntimeError):
    def __init__(self, message='The runtime is not recognized as a BioLib app'):
        self.message = message
        super().__init__(self.message)



import os
import re
from pathlib import Path

import rich.progress
import yaml

from biolib import api, utils
from biolib._internal.data_record.push_data import (
    push_data_path,
    validate_data_path_and_get_files_and_size_of_directory,
)
from biolib._internal.file_utils import get_files_and_size_of_directory, get_iterable_zip_stream
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_api_client.biolib_app_api import BiolibAppApi
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger
from biolib.typing_utils import Iterable, Optional, Set, TypedDict
from biolib.utils.app_uri import parse_app_uri

REGEX_MARKDOWN_INLINE_IMAGE = re.compile(r'!\[(?P<alt>.*)\]\((?P<src>.*)\)')


class DockerProgressDetail(TypedDict):
    current: int
    total: int


class DockerStatusUpdate(TypedDict, total=False):
    status: str
    progressDetail: DockerProgressDetail
    progress: str
    id: str


def process_docker_status_updates(status_updates: Iterable[DockerStatusUpdate], action: str) -> None:
    with rich.progress.Progress() as progress:
        layer_id_to_task_id = {}

        for update in status_updates:
            if 'progressDetail' in update and 'id' in update:
                layer_id = update['id']
                progress_detail = update['progressDetail']

                if layer_id not in layer_id_to_task_id:
                    layer_id_to_task_id[layer_id] = progress.add_task(description=f'[cyan]{action} layer {layer_id}')

                if progress_detail and 'current' in progress_detail and 'total' in progress_detail:
                    progress.update(
                        task_id=layer_id_to_task_id[layer_id],
                        completed=progress_detail['current'],
                        total=progress_detail['total'],
                    )
                elif update['status'] == 'Layer already exists':
                    progress.update(
                        completed=100,
                        task_id=layer_id_to_task_id[layer_id],
                        total=100,
                    )

            elif 'status' not in update and 'progressDetail' not in update:
                print(update)


def set_app_version_as_active(
    app_version_uuid: str,
):
    logger.debug(f'Setting app version {app_version_uuid} as active.')
    api.client.patch(
        path=f'/app_versions/{app_version_uuid}/',
        data={'set_as_active': True},
    )


def push_application(
    app_uri: str,
    app_path: str,
    app_version_to_copy_images_from: Optional[str],
    set_as_active: bool,
    set_as_published: bool,
):
    parsed_uri = parse_app_uri(app_uri)
    app_name = parsed_uri['app_name']

    app_uri_to_fetch = f"@{parsed_uri['resource_name_prefix']}/{parsed_uri['account_handle_normalized']}/{app_name}"

    version = parsed_uri['version']
    semantic_version = f"{version['major']}.{version['minor']}.{version['patch']}" if version else None

    app_path_absolute = Path(app_path).resolve()

    api_client = BiolibApiClient.get()
    if not api_client.is_signed_in:
        # TODO: Create an exception class for expected errors like this that does not print stacktrace

        raise Exception(
            'You must be authenticated to push an application.\n'
            'Please set the environment variable "BIOLIB_TOKEN=[your_deploy_token]"\n'
            f'You can get a deploy key at: {api_client.base_url}/{app_uri_to_fetch}/settings/keys/'
        ) from None

    # prepare zip file
    config_yml_path = app_path_absolute.joinpath('.biolib/config.yml')
    if not config_yml_path.is_file():
        raise BioLibError('The file .biolib/config.yml was not found in the application directory')

    zip_filters: Set[str] = set()
    zip_filters.add('.biolib/config.yml')

    input_files_maps_to_root = False
    app_data_path: Optional[Path] = None
    try:
        with open(config_yml_path) as config_yml_file:
            config = yaml.safe_load(config_yml_file.read())

        app_data = config.get('app_data')
        if app_data:
            if not isinstance(app_data, str):
                raise BioLibError(
                    f'In .biolib/config.yml the value of "app_data" must be a string but got {type(app_data)}'
                )

            app_data_path = app_path_absolute.joinpath(app_data).resolve()
            if not app_data_path.is_dir():
                raise BioLibError(
                    'In .biolib/config.yml the value of "app_data" must be a path to a directory '
                    'in the application directory'
                )

        license_file_relative_path = config.get('license_file', 'LICENSE')
        if app_path_absolute.joinpath(license_file_relative_path).is_file():
            zip_filters.add(license_file_relative_path)

        description_file_relative_path = config.get('description_file', 'README.md')
        description_file_absolute_path = app_path_absolute.joinpath(description_file_relative_path)
        if not description_file_absolute_path.is_file():
            raise BioLibError(f'Could not find {description_file_relative_path}')

        zip_filters.add(description_file_relative_path)
        with open(description_file_absolute_path) as description_file:
            description_file_content = description_file.read()

        for _, img_src_path in re.findall(REGEX_MARKDOWN_INLINE_IMAGE, description_file_content):
            zip_filters.add(img_src_path)

        for _, module in config['modules'].items():
            if module.get('source_files'):
                zip_filters.add('*')

            for mapping in module.get('input_files', []):
                mapping_parts = mapping.split(' ')
                if len(mapping_parts) == 3 and mapping_parts[2] == '/':
                    input_files_maps_to_root = True

    except BioLibError as error:
        raise error from None

    except Exception as error:
        raise BioLibError('Failed to parse the .biolib/config.yml file') from error

    if input_files_maps_to_root:
        logger.error(
            'Error: In your config.yml some module maps input files to "/" (root). '
            'This is potentially an unsafe operation as it allows the user to '
            'overwrite system executables in the module.'
        )
        exit(1)

    files_in_app_dir, _ = get_files_and_size_of_directory(directory=str(app_path_absolute))
    files_to_zip: Set[str] = set()

    for file_path in files_in_app_dir:
        for pattern in zip_filters:
            if pattern == '*' or pattern == file_path:
                files_to_zip.add(file_path)
                break

    original_directory = os.getcwd()
    os.chdir(app_path_absolute.parent)
    files_with_app_dir_prefixed = [f'{app_path_absolute.stem}/{path}' for path in files_to_zip]

    # Workaround as backend currently expects directory objects for root level and .biolib directory
    files_with_app_dir_prefixed.append(f'{app_path_absolute.stem}/')
    files_with_app_dir_prefixed.append(f'{app_path_absolute.stem}/.biolib/')

    byte_iterator = get_iterable_zip_stream(files_with_app_dir_prefixed, chunk_size=50_000_000)
    source_files_zip_bytes = b''.join(byte_iterator)
    os.chdir(original_directory)

    if app_version_to_copy_images_from and app_version_to_copy_images_from != 'active':
        # Get app with `app_version_to_copy_images_from` in app_uri_to_fetch to get the app version public id.
        app_uri_to_fetch += f':{app_version_to_copy_images_from}'

    app_response = BiolibAppApi.get_by_uri(app_uri_to_fetch)
    app = app_response['app']

    if app_data and not app['allow_client_side_execution']:
        raise BioLibError(
            'To push a version with app_data the app must be set to "Allow Client-Side Source Code Access"'
        )
    # push new app version
    new_app_version_json = BiolibAppApi.push_app_version(
        semantic_version=semantic_version,
        app_id=app['public_id'],
        app_name=app['name'],
        author=app['account_handle'],
        set_as_active=False,
        zip_binary=source_files_zip_bytes,
        app_version_id_to_copy_images_from=app_response['app_version']['public_id']
        if app_version_to_copy_images_from
        else None,
    )

    if app_data_path:
        app_data_files_to_zip, app_data_size_in_bytes = validate_data_path_and_get_files_and_size_of_directory(
            data_path=str(app_data_path),
        )
        push_data_path(
            resource_version_uuid=new_app_version_json['public_id'],
            data_path=str(app_data_path),
            data_size_in_bytes=app_data_size_in_bytes,
            files_to_zip=app_data_files_to_zip,
        )

    #  Don't push docker images if copying from another app version
    docker_tags = new_app_version_json.get('docker_tags', {})
    if not app_version_to_copy_images_from and docker_tags:
        logger.info('Found docker images to push.')
        docker_client = BiolibDockerClient.get_docker_client()

        for module_name, repo_and_tag in docker_tags.items():
            docker_image_definition = config['modules'][module_name]['image']
            repo, tag = repo_and_tag.split(':')

            if docker_image_definition.startswith('dockerhub://'):
                docker_image_name = docker_image_definition.replace('dockerhub://', 'docker.io/', 1)
                logger.info(f'Pulling image {docker_image_name} defined on module {module_name} from Dockerhub.')
                dockerhub_repo, dockerhub_tag = docker_image_name.split(':')
                pull_status_updates: Iterable[DockerStatusUpdate] = docker_client.api.pull(
                    decode=True,
                    platform='linux/amd64',
                    repository=dockerhub_repo,
                    stream=True,
                    tag=dockerhub_tag,
                )

                process_docker_status_updates(pull_status_updates, action='Pulling')

            elif docker_image_definition.startswith('local-docker://'):
                docker_image_name = docker_image_definition.replace('local-docker://', '', 1)

            try:
                logger.info(f'Trying to push image {docker_image_name} defined on module {module_name}.')
                image = docker_client.images.get(docker_image_name)
                architecture = image.attrs.get('Architecture')
                if architecture != 'amd64':
                    print(f"Error: '{docker_image_name}' is compiled for {architecture}, expected x86 (amd64).")
                    print('If you are on an ARM processor, try passing --platform linux/amd64 to docker build.')
                    exit(1)
                absolute_repo_uri = f'{utils.BIOLIB_SITE_HOSTNAME}/{repo}'
                image.tag(absolute_repo_uri, tag)

                push_status_updates: Iterable[DockerStatusUpdate] = docker_client.images.push(
                    absolute_repo_uri,
                    tag=tag,
                    stream=True,
                    decode=True,
                    auth_config={
                        'username': 'biolib',
                        # For legacy reasons access token is sent with trailing comma ','
                        'password': api_client.resource_deploy_key or f'{api_client.access_token},',
                    },
                )

                process_docker_status_updates(push_status_updates, action='Pushing')

            except Exception as exception:
                raise BioLibError(f'Failed to tag and push image {docker_image_name}.') from exception

            logger.info(f'Successfully pushed {docker_image_name}')

    app_version_uuid = new_app_version_json['public_id']
    api.client.post(
        path=f'/app-versions/{app_version_uuid}/complete-push/',
        data={'set_as_active': set_as_active, 'set_as_published': set_as_published},
    )

    sematic_version = f"{new_app_version_json['major']}.{new_app_version_json['minor']}.{new_app_version_json['patch']}"
    version_name = 'development ' if not set_as_published else ''
    logger.info(f'Successfully pushed new {version_name}version {sematic_version} of {app_uri}.')

    return {'app_uri': app_uri, 'sematic_version': sematic_version}



import os
import sys

import click

from biolib import templates


@click.command(help='Initialize a project with a .biolib/config.yml file', hidden=True)
def init() -> None:
    cwd = os.getcwd()
    config_file_path = f'{cwd}/.biolib/config.yml'
    readme_path = f'{cwd}/README.md'

    if os.path.exists(config_file_path):
        print(f'The file "{config_file_path}" already exists', file=sys.stderr)
        exit(1)
    else:
        os.makedirs(f'{cwd}/.biolib', exist_ok=True)
        with open(config_file_path, 'w') as config_file:
            config_file.write(templates.example_app.CONFIG_YML)

    if not os.path.exists(readme_path):
        with open(readme_path, 'w') as readme_file:
            project_name = input('Enter a name for your project: ')
            readme_file.write(f'# {project_name}\n')



import json
import re
from pathlib import Path
from typing import Optional

from biolib import api
from biolib._internal.runtime import BioLibRuntimeError, BioLibRuntimeNotRecognizedError, RuntimeJobDataDict
from biolib.biolib_logging import logger
from biolib.typing_utils import cast
from biolib.utils.seq_util import SeqUtil


class Runtime:
    _job_data: Optional[RuntimeJobDataDict] = None

    @staticmethod
    def check_is_environment_biolib_app() -> bool:
        return bool(Runtime._try_to_get_job_data())

    @staticmethod
    def check_is_environment_biolib_cloud() -> bool:
        return Runtime._get_job_data().get('is_environment_biolib_cloud', False)

    @staticmethod
    def get_job_id() -> str:
        return Runtime._get_job_data()['job_uuid']

    @staticmethod
    def get_job_auth_token() -> str:
        return Runtime._get_job_data()['job_auth_token']

    @staticmethod
    def get_job_requested_machine() -> str:
        return Runtime._get_job_data()['job_requested_machine']

    @staticmethod
    def get_app_uri() -> str:
        return Runtime._get_job_data()['app_uri']

    @staticmethod
    def get_secret(secret_name: str) -> bytes:
        assert re.match(
            '^[a-zA-Z0-9_-]*$', secret_name
        ), 'Secret name can only contain alphanumeric characters and dashes or underscores '
        try:
            with open(f'/biolib/secrets/{secret_name}', 'rb') as file:
                return file.read()
        except BaseException as error:
            raise BioLibRuntimeError(f'Unable to get system secret: {secret_name}') from error

    @staticmethod
    def get_temporary_client_secret(secret_name: str) -> bytes:
        assert re.match(
            '^[a-zA-Z0-9_-]*$', secret_name
        ), 'Secret name can only contain alphanumeric characters and dashes or underscores '
        try:
            with open(f'/biolib/temporary-client-secrets/{secret_name}', 'rb') as file:
                return file.read()
        except BaseException as error:
            raise BioLibRuntimeError(f'Unable to get secret: {secret_name}') from error

    @staticmethod
    def set_main_result_prefix(result_prefix: str) -> None:
        job_data = Runtime._get_job_data()
        api.client.patch(
            data={'result_name_prefix': result_prefix},
            headers={'Job-Auth-Token': job_data['job_auth_token']},
            path=f"/jobs/{job_data['job_uuid']}/main_result/",
        )

    @staticmethod
    def set_result_name_prefix(result_prefix: str) -> None:
        Runtime.set_main_result_prefix(result_prefix)

    @staticmethod
    def set_result_name_prefix_from_fasta(path_to_fasta: str) -> None:
        """
        Set BioLib result name to FASTA file name (excluding file extension) or,
        if detecting a BioLib generated name, use ID of the first sequence in the fasta file
        """

        # Set job name to first header if FASTA text input (random BioLib file) detected
        if re.match('^input_[A-Za-z0-9]+.*', path_to_fasta):
            first_id = next(SeqUtil.parse_fasta(path_to_fasta)).id
            result_name = first_id.replace(' ', '_')[:60]
        else:
            result_name = Path(path_to_fasta).stem

        logger.debug(f'Setting result name to "{result_name}" from {path_to_fasta}')
        Runtime.set_result_name_prefix(result_name)

    @staticmethod
    def create_result_note(note: str) -> None:
        job_id = Runtime.get_job_id()
        # Note: Authentication is added by app caller proxy in compute node
        api.client.post(data={'note': note}, path=f'/jobs/{job_id}/notes/')

    @staticmethod
    def _try_to_get_job_data() -> Optional[RuntimeJobDataDict]:
        if not Runtime._job_data:
            try:
                with open('/biolib/secrets/biolib_system_secret') as file:
                    job_data: RuntimeJobDataDict = json.load(file)
            except BaseException:
                return None

            if not job_data['version'].startswith('1.'):
                raise BioLibRuntimeError(f"Unexpected system secret version {job_data['version']} expected 1.x.x")

            Runtime._job_data = job_data

        return cast(RuntimeJobDataDict, Runtime._job_data)

    @staticmethod
    def _get_job_data() -> RuntimeJobDataDict:
        job_data = Runtime._try_to_get_job_data()
        if not job_data:
            raise BioLibRuntimeNotRecognizedError() from None
        return job_data
